/*
 * Copyright (C) 2011 Atmark Techno, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/gpio.h>
#include <linux/i2c.h>
#include <linux/mfd/mc34704/core.h>
#include <linux/rtc/s35390a.h>
#include <mach/common.h>
#include <mach/hardware.h>
#include <mach/i2c.h>
#include "armadillo400_iomux.h"
#include "board.h"
#include "devices.h"

extern struct mc34704_platform_data mc34704_plat;

static struct pad_desc armadillo400_i2c1_pads[] = {
	MX25_PAD_I2C1_CLK__I2C1_CLK(PAD_CTL_PUS_100K_UP | PAD_CTL_ODE),
	MX25_PAD_I2C1_DAT__I2C1_DAT(PAD_CTL_PUS_100K_UP | PAD_CTL_ODE),
};

static struct pad_desc armadillo400_i2c1_gpio_pads[] = {
	MX25_PAD_I2C1_CLK__GPIO_1_12(PAD_CTL_PUS_100K_UP),
	MX25_PAD_I2C1_DAT__GPIO_1_13(PAD_CTL_PUS_100K_UP),
};

static struct pad_desc __maybe_unused armadillo400_i2c2_pads[] = {
#if defined(CONFIG_ARMADILLO400_CON14_3_I2C2_SCL)
	MX25_PAD_GPIO_C__I2C2_SCL(PAD_CTL_PKE | PAD_CTL_PUE | PAD_CTL_PUS_22K_UP | PAD_CTL_ODE),
#endif
#if defined(CONFIG_ARMADILLO400_CON14_4_I2C2_SDA)
	MX25_PAD_GPIO_D__I2C2_SDA(PAD_CTL_PKE | PAD_CTL_PUS_22K_UP | PAD_CTL_ODE),
#endif
};

static struct pad_desc __maybe_unused armadillo400_i2c2_gpio_pads[] = {
#if defined(CONFIG_ARMADILLO400_CON14_3_I2C2_SCL)
	MX25_PAD_GPIO_C__GPIO_C(PAD_CTL_PKE | PAD_CTL_PUE | PAD_CTL_PUS_22K_UP | PAD_CTL_ODE),
#endif
#if defined(CONFIG_ARMADILLO400_CON14_4_I2C2_SDA)
	MX25_PAD_GPIO_D__GPIO_D(PAD_CTL_PKE | PAD_CTL_PUS_22K_UP | PAD_CTL_ODE),
#endif
};

static struct pad_desc __maybe_unused armadillo400_i2c3_pads[] = {
#if defined(CONFIG_ARMADILLO400_CON11_48_I2C3_SCL)
	MX25_PAD_GPIO_A__I2C3_SCL(PAD_CTL_PKE | PAD_CTL_PUE | PAD_CTL_PUS_22K_UP | PAD_CTL_ODE),
#endif
#if defined(CONFIG_ARMADILLO400_CON11_49_I2C3_SDA)
	MX25_PAD_GPIO_B__I2C3_SDA(PAD_CTL_PKE | PAD_CTL_PUE | PAD_CTL_PUS_22K_UP | PAD_CTL_ODE),
#endif
};

static struct pad_desc __maybe_unused armadillo400_i2c3_gpio_pads[] = {
#if defined(CONFIG_ARMADILLO400_CON11_48_I2C3_SCL)
	MX25_PAD_GPIO_A__GPIO_A(PAD_CTL_PKE | PAD_CTL_PUE | PAD_CTL_PUS_22K_UP),
#endif
#if defined(CONFIG_ARMADILLO400_CON11_49_I2C3_SDA)
	MX25_PAD_GPIO_B__GPIO_B(PAD_CTL_PKE | PAD_CTL_PUE | PAD_CTL_PUS_22K_UP),
#endif
};

#define GPIO_I2C_DUMMY_CLOCK_NUM	128
#define GPIO_I2C_DUMMY_CLOCK_FREQ	40000
#define GPIO_I2C_DUMMY_CLOCL_UDELAY	(1000000 / GPIO_I2C_DUMMY_CLOCK_FREQ)

static int gpio_i2c_dummy_clock(unsigned gpio_clk, unsigned gpio_dat)
{
	int i;

	if (gpio_request(gpio_clk, "i2c_clk_gpio"))
		return -EINVAL;

	if (gpio_request(gpio_dat, "i2c_dat_gpio")) {
		gpio_free(gpio_clk);
		return -EINVAL;
	}

	gpio_direction_output(gpio_clk, 0);
	gpio_direction_input(gpio_dat);

	if (gpio_get_value(gpio_dat) == 0) {
		for (i = 0; i < GPIO_I2C_DUMMY_CLOCK_NUM; i++) {
			gpio_set_value(gpio_clk, 1);
			udelay(GPIO_I2C_DUMMY_CLOCL_UDELAY);
			gpio_set_value(gpio_clk, 0);
			udelay(GPIO_I2C_DUMMY_CLOCL_UDELAY);
		}
	}

	gpio_free(gpio_clk);
	gpio_free(gpio_dat);

	return 0;
}

static int platform_i2c_init(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);

	switch (pdev->id) {
	case 0:
		mxc_iomux_v3a_setup_multiple_pads(armadillo400_i2c1_gpio_pads);
		gpio_i2c_dummy_clock(I2C1_CLK_GPIO, I2C1_DAT_GPIO);
		mxc_iomux_v3a_setup_multiple_pads(armadillo400_i2c1_pads);
		break;
	case 1:
#if defined(CONFIG_ARMADILLO400_I2C2_CON14)
		mxc_iomux_v3a_setup_multiple_pads(armadillo400_i2c2_gpio_pads);
		gpio_i2c_dummy_clock(I2C2_CLK_GPIO, I2C2_DAT_GPIO);
		mxc_iomux_v3a_setup_multiple_pads(armadillo400_i2c2_pads);
#endif
		break;
	case 2:
#if defined(CONFIG_ARMADILLO400_I2C3_CON11)
		mxc_iomux_v3a_setup_multiple_pads(armadillo400_i2c3_gpio_pads);
		gpio_i2c_dummy_clock(I2C3_CLK_GPIO, I2C3_DAT_GPIO);
		mxc_iomux_v3a_setup_multiple_pads(armadillo400_i2c3_pads);
#endif
		break;
	default:
		return -ENODEV;
	}

	return 0;
}

static struct imxi2c_platform_data armadillo400_i2c1_data = {
	.bitrate = 40000,
	.init = platform_i2c_init,
};

static struct i2c_board_info armadillo400_i2c1_board_info[] __initdata = {
	{
		.type = "mc34704",
		.addr = 0x54,
		.platform_data = &mc34704_plat,
	},
};

#if defined(CONFIG_ARMADILLO400_SELECTED_I2C2)
#if defined(CONFIG_ARMADILLO400_I2C2_CON14_S35390A)
static struct s35390a_platform_data armadillo400_i2c2_con14_s35390a = {
	.regulator = "s35390a-power",
};
#endif

static struct imxi2c_platform_data armadillo400_i2c2_data = {
	.bitrate = 40000,
	.init = platform_i2c_init,
};

static struct i2c_board_info armadillo400_i2c2_board_info[] __initdata = {
#if defined(CONFIG_ARMADILLO400_I2C2_CON14_S35390A)
	{
		.type = "s35390a",
		.addr = 0x30,
		.platform_data = &armadillo400_i2c2_con14_s35390a,
	},
#endif
#if defined(CONFIG_ARMADILLO400_I2C2_CON14_WM8978)
	{
		.type = "wm8978",
		.addr = 0x1a,
		.platform_data = "wm8978-power",
	},
#endif
};
#endif

#if defined(CONFIG_ARMADILLO400_SELECTED_I2C3)
#if defined(CONFIG_ARMADILLO400_I2C3_CON11_S35390A)
static struct s35390a_platform_data armadillo400_i2c3_con11_s35390a = {
	.regulator = "s35390a-power",
};
#endif

static struct imxi2c_platform_data armadillo400_i2c3_data = {
	.bitrate = 40000,
	.init = platform_i2c_init,
};

static struct i2c_board_info armadillo400_i2c3_board_info[] __initdata = {
#if defined(CONFIG_ARMADILLO400_I2C3_CON11_S35390A)
	{
		.type = "s35390a",
		.addr = 0x30,
		.platform_data = &armadillo400_i2c3_con11_s35390a,
	},
#endif
#if defined(CONFIG_ARMADILLO400_I2C3_CON11_WM8978)
	{
		.type = "wm8978",
		.addr = 0x1a,
		.platform_data = "wm8978-power",
	},
#endif
};
#endif

void __init armadillo400_i2c_init(void)
{
	mxc_register_device(&mxc_i2c_device0,
			    &armadillo400_i2c1_data);
	i2c_register_board_info(mxc_i2c_device0.id,
				armadillo400_i2c1_board_info,
				ARRAY_SIZE(armadillo400_i2c1_board_info));

#if defined(CONFIG_ARMADILLO400_SELECTED_I2C2)
	mxc_register_device(&mxc_i2c_device1,
			    &armadillo400_i2c2_data);
	i2c_register_board_info(mxc_i2c_device1.id,
				armadillo400_i2c2_board_info,
				ARRAY_SIZE(armadillo400_i2c2_board_info));
#endif

#if defined(CONFIG_ARMADILLO400_SELECTED_I2C3)
	mxc_register_device(&mxc_i2c_device2,
			    &armadillo400_i2c3_data);
	i2c_register_board_info(mxc_i2c_device2.id,
				armadillo400_i2c3_board_info,
				ARRAY_SIZE(armadillo400_i2c3_board_info));
#endif
}
