/*
 * armadillo400-wm8978.c
 *   --  Armadillo-400 Driver for Wolfson WM8978 Codec
 *
 * Copyright 2011 Atmark Techno, Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/bitops.h>
#include <linux/platform_device.h>
#include <linux/i2c.h>
#include <linux/err.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>
#include <sound/tlv.h>
#include <sound/initval.h>
#include <mach/armadillo400_wm8978.h>
#include "../codecs/wm8978.h"
#include "imx-ssi.h"
#include "imx-pcm.h"

#define DRV_NAME "armadillo400-wm8978"

#define error(format, arg...) \
	pr_err(DRV_NAME ": " format, ## arg)
#define info(format, arg...) \
	pr_info(DRV_NAME ": " format, ## arg)
#define warning(format, arg...) \
	pr_warning(DRV_NAME ": " format, ## arg)
#define debug(format, arg...) \
	pr_debug(DRV_NAME ": " format, ## arg)

#define ARMADILLO400_WM8978_MCLOCK 24576000 /* actually 24610kHz */

/* WM8978 parameters */
#define WM8978_MCLK_DIV_1	(0 << 5)
#define WM8978_MCLK_DIV_1_5	(1 << 5)
#define WM8978_MCLK_DIV_2	(2 << 5)
#define WM8978_MCLK_DIV_3	(3 << 5)
#define WM8978_MCLK_DIV_4	(4 << 5)
#define WM8978_MCLK_DIV_6	(5 << 5)
#define WM8978_MCLK_DIV_8	(6 << 5)
#define WM8978_MCLK_DIV_12	(7 << 5)

#define WM8978_BCLK_DIV_1	(0 << 2)
#define WM8978_BCLK_DIV_2	(1 << 2)
#define WM8978_BCLK_DIV_4	(2 << 2)
#define WM8978_BCLK_DIV_8	(3 << 2)
#define WM8978_BCLK_DIV_16	(4 << 2)
#define WM8978_BCLK_DIV_32	(5 << 2)

static void armadillo400_dump_codec_reg(struct snd_soc_codec *codec)
{
	int reg;
	u16 val;

	debug("%s()\n", __func__);
	for (reg=0; reg < WM8978_CACHEREGNUM; reg++) {
		val = codec->read(codec, reg);
		debug("\treg 0x%02x(%2d): 0x%03x\n", reg, reg, val);
	}
}

static int armadillo400_hifi_prepare(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_pcm_runtime *runtime = substream->runtime;
	struct snd_soc_device *socdev = rtd->socdev;
	struct snd_soc_codec *codec = socdev->card->codec;

	if (runtime->channels == 1)
		snd_soc_update_bits(codec, WM8978_OUTPUT_CONTROL, 0x40, 0x40);
	else
		snd_soc_update_bits(codec, WM8978_OUTPUT_CONTROL, 0x40, 0);

	return 0;
}

static int armadillo400_hifi_hw_params(struct snd_pcm_substream *substream,
				       struct snd_pcm_hw_params *params)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai_link *pcm_link = rtd->dai;
	struct snd_soc_dai *cpu_dai = pcm_link->cpu_dai;
	struct snd_soc_dai *codec_dai = pcm_link->codec_dai;
	struct imx_ssi *ssi_mode = (struct imx_ssi *)cpu_dai->private_data;
	unsigned int channels = params_channels(params);
	unsigned int rate = params_rate(params);

	u32 dai_format, bclkdiv, mclkdiv, sysclk;
	int ssi_clk_div2, ssi_clk_psr, ssi_clk_pm;
	int ret;

	switch (rate) {
	case 8000:
		bclkdiv = WM8978_BCLK_DIV_4;
		mclkdiv = WM8978_MCLK_DIV_12;
		sysclk = 0;

		ssi_clk_div2 = 0;
		ssi_clk_psr  = 0;
		ssi_clk_pm   = 0x17;
		break;

	case 16000:
		bclkdiv = WM8978_BCLK_DIV_4;
		mclkdiv = WM8978_MCLK_DIV_6;
		sysclk = 0;

		ssi_clk_div2 = 0;
		ssi_clk_psr  = 0;
		ssi_clk_pm   = 0x0b;
		break;

	case 32000:
		bclkdiv = WM8978_BCLK_DIV_4;
		mclkdiv = WM8978_MCLK_DIV_3;
		sysclk = 0;

		ssi_clk_div2 = 0;
		ssi_clk_psr  = 0;
		ssi_clk_pm   = 0x05;
		break;

	case 48000:
		bclkdiv = WM8978_BCLK_DIV_4;
		mclkdiv = WM8978_MCLK_DIV_2;
		sysclk = 0;

		ssi_clk_div2 = 0;
		ssi_clk_psr  = 0;
		ssi_clk_pm   = 0x03;
		break;

	default:
		error("unsupported format\n");
		return -EINVAL;
	}

	dai_format = SND_SOC_DAIFMT_I2S | SND_SOC_DAIFMT_NB_NF |
		SND_SOC_DAIFMT_CBS_CFS;

	ssi_mode->sync_mode = 1;
	ssi_mode->network_mode = 1;

	/* set codec DAI configuration */
	ret = snd_soc_dai_set_fmt(codec_dai, dai_format);
	if (ret < 0) {
		error("codec set_fmt failed\n");
		return ret;
	}

	/* set cpu DAI configuration */
	ret = snd_soc_dai_set_fmt(cpu_dai, dai_format);
	if (ret < 0) {
		error("cpu set_fmt failed\n");
		return ret;
	}

	/* set i.MX active slot mask */
	ret = snd_soc_dai_set_tdm_slot(cpu_dai,
				       channels == 1 ? 0xfffffffe : 0xfffffffc,
				       channels == 1 ? 0xfffffffe : 0xfffffffc,
				       2, 32);
	if (ret < 0) {
		error("cpu set_tdm_slot failed\n");
		return ret;
	}

	ret = snd_soc_dai_set_sysclk(cpu_dai, IMX_SSP_SYS_CLK,
				     0, SND_SOC_CLOCK_OUT);
	if (ret < 0) {
		error("cpu set_sysclk failed\n");
		return ret;
	}

	/* codec PLL input is rate from MCLK */
	ret = snd_soc_dai_set_sysclk(codec_dai, WM8978_MCLK,
				     ARMADILLO400_WM8978_MCLOCK,
				     SND_SOC_CLOCK_IN);
	if (ret < 0) {
		error("codec set_pll failed\n");
		return ret;
	}

	snd_soc_dai_set_clkdiv(cpu_dai, IMX_SSI_TX_DIV_2, ssi_clk_div2);
	snd_soc_dai_set_clkdiv(cpu_dai, IMX_SSI_RX_DIV_2, ssi_clk_div2);

	snd_soc_dai_set_clkdiv(cpu_dai, IMX_SSI_TX_DIV_PSR, ssi_clk_psr);
	snd_soc_dai_set_clkdiv(cpu_dai, IMX_SSI_RX_DIV_PSR, ssi_clk_psr);

	snd_soc_dai_set_clkdiv(cpu_dai, IMX_SSI_TX_DIV_PM, ssi_clk_pm);
	snd_soc_dai_set_clkdiv(cpu_dai, IMX_SSI_RX_DIV_PM, ssi_clk_pm);

	armadillo400_dump_codec_reg(codec_dai->codec);

	return 0;
}

static const DECLARE_TLV_DB_SCALE(inpga_tlv, -1200, 75, 0);
static const struct snd_kcontrol_new wm8978_snd_ext_controls[] = {
	SOC_DOUBLE_R_TLV("Mic Volume",
		WM8978_LEFT_INP_PGA_CONTROL, WM8978_RIGHT_INP_PGA_CONTROL,
		0, 63, 0, inpga_tlv),
};

static const struct snd_soc_dapm_widget armadillo400_dapm_widgets[] = {
	SND_SOC_DAPM_HP("Headphone Jack", NULL),
	SND_SOC_DAPM_SPK("Speaker", NULL),
	SND_SOC_DAPM_MIC("Mic Jack", NULL),
};

static const struct snd_soc_dapm_route audio_map[] = {
	/* Headphone output connected to LHP/RHP, enable OUT4 for VMID */
	{ "Headphone Jack", NULL, "LHP" },
	{ "Headphone Jack", NULL, "RHP" },

	/* speaker connected to LSPK/RSPK */
	{ "Speaker", NULL, "LSPK" },
	{ "Speaker", NULL, "RSPK" },

	/* mic is connected to LMICP/N */
	{ "LMICP", NULL, "Mic Jack" },
	{ "LMICN", NULL, "Mic Jack" },
};

static void armadillo400_wm8978_codec_init(struct snd_soc_codec *codec)
{
	snd_soc_write(codec, WM8978_ADDITIONAL_CONTROL, 0x1);
	snd_soc_write(codec, WM8978_JACK_DETECT_CONTROL_1, 0x050);
	snd_soc_write(codec, WM8978_JACK_DETECT_CONTROL_2, 0x012);
}

static int armadillo400_wm8978_dai_init(struct snd_soc_codec *codec)
{
	int ret;

	armadillo400_wm8978_codec_init(codec);

	/* Add armadillo500fx specific widgets */
	ret = snd_soc_dapm_new_controls(codec, armadillo400_dapm_widgets,
					ARRAY_SIZE(armadillo400_dapm_widgets));
	if (ret < 0)
		error("widgets ret: %d\n", ret);

	/* set up armadillo400 audio specific path audio_map */
	ret = snd_soc_dapm_add_routes(codec, audio_map, ARRAY_SIZE(audio_map));
	if (ret < 0)
		return ret;

	/* set up wm8978 external controls for armadillo */
	snd_soc_add_controls(codec, wm8978_snd_ext_controls,
			     ARRAY_SIZE(wm8978_snd_ext_controls));

	/* set up force enable pins */
	snd_soc_dapm_force_enable_pin(codec, "Left DAC");
	snd_soc_dapm_force_enable_pin(codec, "Left Output Mixer");
	snd_soc_dapm_force_enable_pin(codec, "Right DAC");
	snd_soc_dapm_force_enable_pin(codec, "Right Output Mixer");

	snd_soc_dapm_force_enable_pin(codec, "Left ADC");
	snd_soc_dapm_force_enable_pin(codec, "Left Input Mixer");

	/* set up NC codec pins */
	snd_soc_dapm_nc_pin(codec, "RMICP");
	snd_soc_dapm_nc_pin(codec, "RMICN");
	snd_soc_dapm_nc_pin(codec, "LAUX");
	snd_soc_dapm_nc_pin(codec, "RAUX");

	snd_soc_dapm_sync(codec);

	return 0;
}

static struct snd_soc_ops armadillo400_hifi_ops = {
	.hw_params = armadillo400_hifi_hw_params,
	.prepare = armadillo400_hifi_prepare,
};

static struct snd_soc_dai_link armadillo400_wm8978_dai_link = {
	.name = "wm8978",
	.stream_name = "WM8978",
	.codec_dai = &wm8978_dai,
	.init = armadillo400_wm8978_dai_init,
	.ops = &armadillo400_hifi_ops,
};

static struct snd_soc_card snd_soc_card_armadillo400 = {
	.name = "armadillo400",
	.platform = &imx_soc_platform,
	.dai_link = &armadillo400_wm8978_dai_link,
	.num_links = 1,
};

static struct snd_soc_device armadillo400_wm8978_snd_soc_dev = {
	.card = &snd_soc_card_armadillo400,
	.codec_dev = &soc_codec_dev_wm8978,
};

static int __devinit armadillo400_wm8978_probe(struct platform_device *pdev)
{
	struct armadillo400_wm8978_audio_data *priv;
	int ret;

	priv = pdev->dev.platform_data;
	if (!priv)
		return -ENODEV;

	priv->soc_audio_dev = platform_device_alloc("soc-audio", pdev->id);
	if (!priv->soc_audio_dev)
		return -ENOMEM;

	platform_set_drvdata(priv->soc_audio_dev,
			     &armadillo400_wm8978_snd_soc_dev);
	armadillo400_wm8978_snd_soc_dev.dev = &priv->soc_audio_dev->dev;

	if (priv->init)
		priv->init(pdev);

	armadillo400_wm8978_dai_link.cpu_dai = imx_ssi_dai[pdev->id * 2];

	ret = platform_device_add(priv->soc_audio_dev);
	if (ret)
		platform_device_put(priv->soc_audio_dev);

	return 0;
}

static int __devexit armadillo400_wm8978_remove(struct platform_device *pdev)
{
	struct armadillo400_wm8978_audio_data *priv;

	priv = pdev->dev.platform_data;
	platform_device_unregister(priv->soc_audio_dev);

	return 0;
}

static struct platform_driver armadillo400_wm8978_driver = {
	.driver = {
		.name = "armadillo400-wm8978-audio",
		.owner = THIS_MODULE,
	},
	.probe = armadillo400_wm8978_probe,
	.remove = __devexit_p(armadillo400_wm8978_remove),
};

static int __init armadillo400_wm8978_init(void)
{
	return platform_driver_register(&armadillo400_wm8978_driver);
}
module_init(armadillo400_wm8978_init);

static void __exit armadillo400_wm8978_exit(void)
{
	platform_driver_unregister(&armadillo400_wm8978_driver);
}
module_exit(armadillo400_wm8978_exit);
